import json, time, sys, os, shutil
from argparse import ArgumentParser, Namespace, RawDescriptionHelpFormatter
import argparse

from bscampp import get_logger, log_exception, __version__
from bscampp.configs import *
from bscampp.functions import *
import bscampp.utils as utils

from multiprocessing import Manager
from concurrent.futures import ProcessPoolExecutor

_LOG = get_logger(__name__)

# process pool initializer
def initial_pool(parser, cmdline_args):
    # avoid redundant logging for child process
    buildConfigs(parser, cmdline_args, child_process=True)

# main pipeline for BSCAMPP
def bscampp_pipeline(*args, **kwargs):
    t0 = time.perf_counter()
    m = Manager(); lock = m.Lock()

    # set up a dry run if specified
    dry_run = False
    if 'dry_run' in kwargs and isinstance(kwargs['dry_run'], bool):
        dry_run = kwargs['dry_run']

    # parse command line arguments and build configurations
    parser, cmdline_args = parseArguments(dry_run=dry_run)

    # initialize multiprocessing (if needed)
    _LOG.warning('Initializing ProcessPoolExecutor...')
    # maximally concurrently run Configs.num_cpus // 2 jobs, each job
    # can use 2 threads
    pool = ProcessPoolExecutor(Configs.max_workers, initializer=initial_pool,
            initargs=(parser, cmdline_args,))

    # (0) temporary files wrote to here
    if not dry_run:
        workdir = os.path.join(Configs.outdir, f'tmp{Configs.tmpfilenbr}')
        try:
            if not os.path.isdir(workdir):
                os.makedirs(workdir)
        except OSError:
            log_exception(_LOG)
    else:
        workdir = os.getcwd()

    # (1) read in tree, alignment, and separate reference sequences from
    # query sequences
    tree, leaf_dict, aln_path, aln, qaln_path, qaln, qname_map, qname_map_rev \
            = readData(workdir, dry_run=dry_run)

    # (2) compute closest leaves for all query sequences
    query_votes_dict, query_top_vote_dict = getClosestLeaves(
            aln_path, qaln_path, aln, qaln, workdir, dry_run=dry_run)

    # (3) first assign all queries to their closest-leaf subtrees,
    # then do reassignment to minimize distance between each's top vote
    # and the subtree's seed leaf
    new_subtree_dict, placed_query_list = assignQueriesToSubtrees(
            query_votes_dict, query_top_vote_dict, tree, leaf_dict,
            dry_run=dry_run)

    # (4) perform placement for each subtree
    output_jplace = placeQueriesToSubtrees(tree, leaf_dict, new_subtree_dict,
            placed_query_list, aln, qaln, cmdline_args, workdir,
            qname_map, qname_map_rev,
            pool, lock, dry_run=dry_run)

    # (5) write the output jplace to local
    writeOutputJplace(output_jplace, dry_run=dry_run)

    # shutdown pool
    _LOG.warning('Shutting down ProcessPoolExecutor...')
    pool.shutdown()
    _LOG.warning('ProcessPoolExecutor shut down.')
    
    # clean up temp files if not keeping
    if not Configs.keeptemp:
        _LOG.info('Removing temporary files...')
        clean_temp_files()

    # stop BSCAMPP 
    send = time.perf_counter()
    _LOG.info('BSCAMPP completed in {} seconds...'.format(send - t0))

    if dry_run:
        return True
    else:
        return False


# main pipeline for SCAMPP
def scampp_pipeline(*args, **kwargs):
    t0 = time.perf_counter()
    m = Manager(); lock = m.Lock()

    # set up a dry run if specified
    dry_run = False
    if 'dry_run' in kwargs and isinstance(kwargs['dry_run'], bool):
        dry_run = kwargs['dry_run']

    # parse command line arguments and build configurations
    parser, cmdline_args = parseArguments(dry_run=dry_run, method="SCAMPP")

    # initialize multiprocessing (if needed)
    _LOG.warning('Initializing ProcessPoolExecutor...')
    pool = ProcessPoolExecutor(Configs.num_cpus, initializer=initial_pool,
            initargs=(parser, cmdline_args,))

    # (0) temporary files wrote to here
    if not dry_run:
        workdir = os.path.join(Configs.outdir, f'tmp{Configs.tmpfilenbr}')
        try:
            if not os.path.isdir(workdir):
                os.makedirs(workdir)
        except OSError:
            log_exception(_LOG)
    else:
        workdir = os.getcwd()

    # (1) read in tree, alignment, and separate reference sequences from
    # query sequences
    tree, leaf_dict, aln_path, aln, qaln_path, qaln, qname_map, qname_map_rev \
            = readData(workdir, dry_run=dry_run)

    # (2) compute closest leaves for all query sequences
    query_votes_dict, query_top_vote_dict = getClosestLeaves(
            aln_path, qaln_path, aln, qaln, workdir, dry_run=dry_run)

    # (3) first assign each query to the subtree built using the closest 
    # leaf as the seed sequence 
    new_subtree_dict, placed_query_list = buildQuerySubtrees(
            query_votes_dict, query_top_vote_dict, tree, leaf_dict,
            dry_run=dry_run)

    # (4) perform placement for each subtree
    output_jplace = placeQueriesToSubtrees(tree, leaf_dict, new_subtree_dict,
            placed_query_list, aln, qaln, cmdline_args, workdir,
            qname_map, qname_map_rev,
            pool, lock, dry_run=dry_run)

    # (5) write the output jplace to local
    writeOutputJplace(output_jplace, dry_run=dry_run)

    # shutdown pool
    _LOG.warning('Shutting down ProcessPoolExecutor...')
    pool.shutdown()
    _LOG.warning('ProcessPoolExecutor shut down.')
    
    # clean up temp files if not keeping
    if not Configs.keeptemp:
        _LOG.info('Removing temporary files...')
        clean_temp_files()

    # stop SCAMPP 
    send = time.perf_counter()
    _LOG.info('SCAMPP completed in {} seconds...'.format(send - t0))

    if dry_run:
        return True
    else:
        return False


def clean_temp_files():
    # all temporary files/directories to remove
    temp_items = [f'tmp{Configs.tmpfilenbr}']
    for temp in temp_items:
        temp_path = os.path.join(Configs.outdir, temp)
        if os.path.isfile(temp_path):
            os.remove(temp_path)
        elif os.path.isdir(temp_path):
            shutil.rmtree(temp_path)
        else:
            continue
        _LOG.info(f'- Removed {temp}')

def parseArguments(dry_run=False, method="BSCAMPP"):
    default_outdir = f"{method.lower()}_output"
    default_outname = f"{method.lower()}_result"

    parser = _init_parser(default_outdir=default_outdir,
            default_outname=default_outname)
    cmdline_args = sys.argv[1:]

    if dry_run:
        cmdline_args = ['-i', 'dummy.info', '-t', 'dummy.tre',
                '-a', 'dummy.fa'] 
    
    # build config
    buildConfigs(parser, cmdline_args)
    _LOG.info('{} is running with: {}'.format(method,
        ' '.join(cmdline_args)))
    getConfigs()

    return parser, cmdline_args

def _init_parser(default_outdir="bscampp_output",
        default_outname="bscampp_result"):
    # example usage
    example_usages = '''Example usages:
> (1) Default
    %(prog)s -i raxml.bestModel -t reference.tre -a alignment.fa  
> (2) Separate alignment file for query sequences
    %(prog)s -i raxml.bestModel -t reference.tre -a reference.fa -q query.fa
> (3) Use pplacer instead of EPA-ng as base method (need RAxML-ng info or FastTree log file)
    %(prog)s -i fasttree.log -t reference.tre -a alignment.fa --placement-method pplacer
'''

    parser = ArgumentParser(
            description=(
                "This program runs BSCAMPP/SCAMPP, a scalable phylogenetic "
                "placement framework that scales EPA-ng/pplacer "
                "to very large tree placement."
                ),
            conflict_handler='resolve',
            epilog=example_usages,
            formatter_class=utils.SmartHelpFormatter,
            )
    parser.add_argument('-v', '--version', action='version',
            version="%(prog)s " + __version__)
    parser.groups = dict()
    required = True

    ## add a subcommand for updating configuration file without running
    ## the BSCAMPP pipeline
    #subparsers = parser.add_subparsers(dest='command',
    #        help='Subcommands for BSCAMPP')
    #update_parser = subparsers.add_parser('update-configs',
    #        help='Update the configuration file without running BSCAMPP.')

    ## try update args requirement if subcommand(s) are used
    #if 'update-configs' in sys.argv:
    #    required = False

    # basic group
    basic_group = parser.add_argument_group(
            "Basic parameters".upper(),
            "These are the basic parameters for BSCAMPP/SCAMPP.")
    parser.groups['basic_group'] = basic_group

    basic_group.add_argument('--placement-method', type=str,
                  help='The base placement method to use. Default: epa-ng',
                  choices=['epa-ng', 'pplacer'], default='epa-ng',
                  required=False)
    basic_group.add_argument("-i", "--info", "--info-path", type=str,
                  dest="info_path",
                  help=("Path to model parameters. E.g., .bestModel "
                  "from RAxML/RAxML-ng"),
                  required=required, default=None)
    basic_group.add_argument("-t", "--tree", "--tree-path", type=str,
                  dest="tree_path",
                  help="Path to reference tree with estimated branch lengths",
                  required=required, default=None)
    basic_group.add_argument("-a", "--alignment", "--aln-path", type=str,
                  dest="aln_path",
                  help=("Path for reference sequence alignment in "
                  "FASTA format (can be a .gz file). "
                  "Optionally with query sequences. "
                  "Query alignment can be specified with --qaln-path"), 
                  required=required, default=None)
    basic_group.add_argument("-q", "--qalignment", "--qaln-path", type=str,
                  dest="qaln_path",
                  help=("Optionally provide path to query sequence alignment "
                  "in FASTA format (can be a .gz file). Default: None"),
                  required=False, default=None)
    #basic_group.add_argument("--molecule", type=str,
    #              choices=['nucl', 'nucleotide', 'prot', 'protein'],
    #              help=("Specify nucleotide or protein sequences. "
    #              "Default: infer datatype"),
    #              required=False, default=None)
    basic_group.add_argument("-d", "--outdir", type=str,
                  help="Directory path for output. Default: bscampp_output/",
                  required=False, default=default_outdir)
    basic_group.add_argument("-o", "--output", type=str, dest="outname",
                  help="Output file name. Default: bscampp_result.jplace",
                  required=False, default=f"{default_outname}.jplace")
    basic_group.add_argument("--threads", "--num-cpus", type=int,
                  dest="num_cpus",
                  help="Number of cores for parallelization, default: -1 (all)",
                  required=False, default=-1)
    basic_group.add_argument("--cpus-per-job", type=int,
                  help="Number of cores to use for each job, default: 2",
                  required=False, default=2)

    # advanced parameter settings
    advance_group = parser.add_argument_group(
            "Advance parameters".upper(),
            ("These parameters control how BSCAMPP is run. "
             "The default values are set based on experiments."
             ))
    parser.groups['advance_group'] = advance_group

    #advance_group.add_argument("-m", "--model", type=str,
    #              help=("Model used for edge distances. EPA-ng will use the "
    #              "provided info_path (*.bestModel) for model. "
    #              "Default: GTR for nucleotide, LG for protein"),
    #              required=False, default=None)
    advance_group.add_argument("-b", "--subtreesize", type=int,
                  help="Integer size of the subtree. Default: 2000",
                  required=False, default=2000)
    advance_group.add_argument("-V", "--votes", type=int,
                  help="(BSCAMPP only) Number of votes per query sequence. "
                  "Default: 5",
                  required=False, default=5)
    advance_group.add_argument("--subtreetype", type=str,
                  help="(SCAMPP only) Options for collecting "
                  "nodes for the subtree - d for edge weighted "
                  "distances, n for node distances, h for Hamming "
                  "distances. Default: d",
                  required=False, default='d')
    advance_group.add_argument("--similarityflag", type=str2bool,
                  help="Boolean, True if maximizing sequence similarity "
                  "instead of simple Hamming distance (ignoring gap "
                  "sites in the query). Default: True",
                  required=False, default=True)
    
    # miscellaneous group
    misc_group = parser.add_argument_group(
            "Miscellaneous parameters".upper(),)
    parser.groups['misc_group'] = misc_group

    misc_group.add_argument("-n","--tmpfilenbr", type=int,
                  help="Temporary file indexing (e.g., tmp0/). Default: 0",
                  required=False, default=0)
    misc_group.add_argument("--fragmentflag", type=str2bool,
                  help="If queries contains fragments. Does not do anything "
                  "if similarity flag is set to True. Default: True",
                  required=False, default=True)
    misc_group.add_argument("--keeptemp", type=str2bool,
                  help="Boolean, True to keep all temporary files. "
                  "Default: False",
                  required=False, default=False)
    return parser

def str2bool(b):
    if isinstance(b, bool):
       return b
    if b.lower() in ('yes', 'true', 't', 'y', '1'):
        return True
    elif b.lower() in ('no', 'false', 'f', 'n', '0'):
        return False
    else:
        raise argparse.ArgumentTypeError('Boolean value expected.')        



import os, sys, shutil
try:
    import configparser
except ImportError:
    import ConfigParser as configparser
from argparse import ArgumentParser, Namespace
from platform import platform

def find_main_config(homepath):
    with open(homepath, 'r') as f:
        _root_dir = f.read().strip()
        main_config_path = os.path.join(_root_dir, 'main.config')
        if os.path.exists(main_config_path):
            return _root_dir, main_config_path
        else:
            return None, None

'''
Initialize the config file at ~/.bscampp/main.config
By default will prioritize existing software from the user environment
'''
def init_config_file(homepath, rerun=False, prioritize_user_software=True):
    if not rerun:
        # make sure home.path exists
        if not os.path.exists(homepath):
            print(f'Cannot find home.path: {homepath}, regenerating...')
        else:
            _root_dir, main_config_path = find_main_config(homepath)
            if _root_dir is not None:
                return _root_dir, main_config_path
            else:
                print(f'Cannot find main.config, regenerating...')

    _root_dir = os.path.expanduser('~/.bscampp')
    main_config_path = os.path.join(_root_dir, 'main.config')
    print(f'Initializing the config file at: {main_config_path}')

    # write to local home.path and _root_dir
    if not os.path.isdir(_root_dir):
        os.mkdir(_root_dir)
    with open(homepath, 'w') as f:
        f.write(_root_dir)

    # create main.config based on the default.config at this file's location
    _config_path = os.path.join(os.path.dirname(__file__), 'default.config')
    cparser = configparser.ConfigParser()
    cparser.optionxform = str
    assert os.path.exists(_config_path), \
            'default config file missing! Please redownload from GitHub\n'

    if os.path.exists(main_config_path):
        print(f'Main configuration file {main_config_path} exists...')
        print('Overwriting the existing config file...')
    print('\n')

    with open(_config_path, 'r') as f:
        cparser.read_file(f)

    # check platform, e.g., macOS or linux, etc.
    platform_name = platform()
    print(f'System is: {platform_name}')

    tools_dir = os.path.join(os.path.dirname(__file__), 'tools')
    set_sections = ['basic']

    # default path to all potential binaries
    cparser.set('basic', 'pplacer_path',
            os.path.join(tools_dir, 'pplacer'))
    cparser.set('basic', 'epang_path',
            os.path.join(tools_dir, 'epa-ng'))
    cparser.set('basic', 'hamming_distance_dir',
            os.path.join(tools_dir, 'hamming_distance'))
    
    # prioritize user's software
    if prioritize_user_software:
        print('Detecting existing software from user\'s environment...')
        softwares = ['pplacer', 'epa-ng', 'taxit']
        for software in softwares:
            sname = software.replace('-', '')
            software_path = shutil.which(software)
            if software_path:
                print('\t{}: {}'.format(software, software_path)) 
                cparser.set('basic', f'{sname}_path', software_path)
    with open(main_config_path, 'w') as f:
        cparser.write(f)
    print(f'\n(Done) main.config was written to: {main_config_path}') 
    print(f'If you want to make changes, please directly edit {main_config_path}') 
    return _root_dir, main_config_path



import json, time, os, sys
import treeswift
from collections import defaultdict, Counter
import subprocess

from bscampp import get_logger, log_exception
from bscampp.configs import Configs
from bscampp.jobs import GenericJob, EPAngJob, TaxtasticJob, PplacerTaxtasticJob
from bscampp.utils import write_fasta
import bscampp.utils as utils

import concurrent.futures

# suppress userwarning when doing subtree suppress_unifurcations
import warnings
warnings.filterwarnings("ignore", category=UserWarning)

_LOG = get_logger(__name__)

############################# helper functions ################################
'''
Function to recompile binaries from the given directory.
Assumption, the directory contains a CMakeLists.txt file
'''
def recompileBinariesFromDir(dir):
    _LOG.warning(f"Recompiling binaries with cmake/make at {dir}")

    # need to recompile the binaries
    cmake_p = subprocess.Popen(['cmake', dir],
            cwd=dir, stdout=subprocess.PIPE,
            stderr=subprocess.PIPE, text=True)
    cmake_stdout, cmake_stderr = cmake_p.communicate()

    if cmake_p.returncode != 0:
        _LOG.error("cmake failed!")
        print("STDOUT:", cmake_stdout)
        print("STDERR:", cmake_stderr)
        exit(cmake_p.returncode)
    else:
        _LOG.warning("cmake succeeded!")

    # run make
    make_p = subprocess.Popen(['make'],
            cwd=dir, stdout=subprocess.PIPE,
            stderr=subprocess.PIPE, text=True)
    make_stdout, make_stderr = make_p.communicate()
    
    if make_p.returncode != 0:
        _LOG.error(f"make failed!")
        exit(make_p.returncode)
    else:
        _LOG.warning("make succeeded!")
    _LOG.warning(f"Successfully recompiled binaries at {dir}!")

'''
Function to check hamming/fragment_hamming/homology binaries are executable, 
since they were compiled using dynamic library
'''
def ensureBinaryExecutable(binpath):
    dir = os.path.dirname(binpath)

    # binpath does not exist
    b_recompile = False
    if not os.path.exists(binpath):
        _LOG.warning(f"{binpath} does not exist!")
        b_recompile = True
    else:
        """
            added @ 6.13.2025 by Chengze Shen
            - try-catch OSError to indicate that the binary files
            - are not executable on the current sytem and need to be
            - recompiled
        """
        try:
            p = subprocess.Popen([binpath], stdout=subprocess.PIPE,
                    stderr=subprocess.PIPE)
            stdout, stderr = p.communicate()
            returncode = p.returncode
        except OSError as e:
            # indicating we need to recompile: anything other than 255 or -1
            returncode = 7

        # 255 or -1 indicates that the binaries work
        if returncode == 255 or returncode == -1:
            pass
        else:
            _LOG.warning(f"{binpath} return code is {returncode}!")
            b_recompile = True

    if b_recompile:
        recompileBinariesFromDir(dir)
    return

########################## end of helper functions ############################

'''
Function to read in the placement tree and alignment.
If query alignment is provided, will use the provided query instead of
the ones (potentially) included in the reference alignment
'''
def readData(workdir, dry_run=False):
    t0 = time.perf_counter()
    _LOG.info('Reading in input data...')

    if dry_run:
        return None, dict(), '', dict(), '', dict(), dict(), dict()

    # (1) load reference tree
    tree = treeswift.read_tree_newick(Configs.tree_path)
    tree.resolve_polytomies()

    leaf_dict = tree.label_to_node(selection='leaves')
    # clean the leaf keys so that ' or " are not present
    ori_keys = list(leaf_dict.keys())
    for key in ori_keys:
        _node = leaf_dict[key]
        new_key = key.replace('\'', '')
        new_key = new_key.replace('\"', '')
        leaf_dict.pop(key)
        leaf_dict[new_key] = _node

    # (2) load reference alignment and query alignment (if provided) 
    if Configs.qaln_path is not None: 
        ref_dict = utils.read_data(Configs.aln_path)
        q_dict = utils.read_data(Configs.qaln_path)
        #aln_path, qaln_path = Configs.aln_path, Configs.qaln_path
    else:
        aln_dict = utils.read_data(Configs.aln_path)
        ref_dict, q_dict = utils.seperate(aln_dict, leaf_dict)

    # Added on 3.8.2025 by Chengze Shen
    #   - to ensure that any characters from the query has correct names
    #     (e.g., having ":" can cause trouble), have a qname_map that maps
    #     each taxon name to an idx
    qidx = 1
    qname_map = dict()
    qname_map_rev = dict()
    for name in q_dict.keys():
        cvt = str(qidx).zfill(16)   # 16 digits
        qname_map[name] = cvt
        qname_map_rev[cvt] = name
        qidx += 1
    # modify q_dict as well
    for name, cvt in qname_map.items():
        q_dict[cvt] = q_dict[name]
        q_dict.pop(name)

    # after separating queries from the reference alignment, write
    # them to to TEMP/
    # Updated on 3.5.2025 by Chengze Shen
    #   - regardless of the input choices, write a copy of both reference
    #     and query alignment to the workdir
    qaln_path = os.path.join(workdir, 'qaln.fa')
    write_fasta(qaln_path, q_dict)
    
    aln_path = os.path.join(workdir, 'aln.fa')
    write_fasta(aln_path, ref_dict)


    t1 = time.perf_counter()
    _LOG.info('Time to read in input data: {} seconds'.format(t1 - t0))
    return tree, leaf_dict, aln_path, ref_dict, qaln_path, q_dict, \
            qname_map, qname_map_rev

'''
Function to get the closest leaf for each query sequence based on Hamming
distance
'''
def getClosestLeaves(aln_path, qaln_path, aln, qaln, workdir, dry_run=False):
    t0 = time.perf_counter()
    _LOG.info('Computing closest leaves for query sequences...')

    if dry_run:
        return dict(), dict()

    query_votes_dict = dict()
    query_top_vote_dict = dict()
    tmp_output = os.path.join(workdir, 'closest.txt') 
    
    if Configs.subtreetype == "h":
        Configs.votes = Configs.subtreesize

    if Configs.similarityflag:
        job_type = 'homology'
    else:
        if Configs.fragmentflag:
            job_type = 'fragment_hamming'
        else: 
            job_type = 'hamming'
    binpath = os.path.join(Configs.hamming_distance_dir, job_type)
    cmd = [binpath]

    # Added @ 3.9.2025 by Chengze Shen
    #   - check if binpath is executable, since the compiled files use dynamic 
    #     libraries.
    #     If works: should have return code 255
    #     If not: should have return code 1,
    #             recompile the binaries using cmake and make
    ensureBinaryExecutable(binpath)

    cmd.extend([aln_path, str(len(aln)), qaln_path, str(len(qaln)),
        tmp_output, str(Configs.votes)])
    job = GenericJob(cmd=cmd, job_type=job_type)
    _ = job.run()
    #os.system(' '.join(cmd))

    # process closest leaves 
    unusable_queries = set()
    f = open(tmp_output)
    for line in f:
        line = line.strip()
        y = line.split(',')
        name = y.pop(0)
        for idx, taxon in enumerate(y):
            leaf, hamming = taxon.split(':')
            y[idx] = (leaf, int(hamming))

        y = sorted(y, key=lambda x: x[1])
        for idx, taxon in enumerate(y):
            y[idx] = taxon[0]

        if name.find(':') >= 0:
            name_list = name.split(":")
            name = name_list[0]
            ungapped_length = name_list[1]
            if y[0] == ungapped_length:
                _LOG.warning(f'Sequence {name}: no homologous sites found, '
                        'removed before placement.')
                unusable_queries.add(name)
        if name not in unusable_queries:
            query_votes_dict[name] = y
            query_top_vote_dict[name] = y[0]
    f.close()
    
    t1 = time.perf_counter()
    _LOG.info('Time to compute closest leaves: {} seconds'.format(t1 - t0)) 
    return query_votes_dict, query_top_vote_dict

'''
Function to assign queries to subtrees based on their votes
'''
def assignQueriesToSubtrees(query_votes_dict, query_top_vote_dict,
        tree, leaf_dict, dry_run=False):
    t0 = time.perf_counter()
    _LOG.info('Adding query votes to the placement tree...')

    if dry_run:
        return dict(), []

    # (1) go over the query votes and add them to corresponding leaves
    lf_votes = Counter()
    leaf_queries = dict()
    for name, y in query_votes_dict.items():
        lf_votes.update(y)
        for ind, leaf in enumerate(y):
            top_vote = False
            if ind == 0:
                top_vote = True
            if leaf not in leaf_queries:           
                leaf_queries[leaf] = {(name,top_vote)}
            else:
                leaf_queries[leaf].add((name,top_vote))

    subtree_dict = dict()
    subtree_leaf_label_dict = dict()
    most_common_index = 0
    
    # assign queries to subtrees, and remove them from the pool
    # repeat until all queries are assigned
    while len(query_votes_dict) > 0:
        _LOG.info("queries left to assign: {}".format(len(query_votes_dict)))
        (seed_label, node_votes) = lf_votes.most_common(
                most_common_index+1)[most_common_index]
        
        node_y = leaf_dict[seed_label]
        # extract [subtreesize] leaves
        labels = utils.subtree_nodes_with_edge_length(tree, node_y,
                Configs.subtreesize)
        subtree = tree.extract_tree_with(labels)
        label_set = set(labels)

        queries_by_subtree = set()
        subtree_query_set = set()

        # gather any other queries that can be used with this subtree
        for label in labels:
            leaf_queries_remove_set = set()
            if label in leaf_queries:
                    
                for leaf_query, top_vote in leaf_queries[label]:
                
                    if leaf_query not in query_votes_dict:
                        leaf_queries_remove_set.add((leaf_query, top_vote))
                        continue
                        
                    if top_vote:
                        subtree_query_set.add(leaf_query)
                        leaf_queries_remove_set.add((leaf_query, top_vote))
                    
                leaf_queries[label].difference_update(leaf_queries_remove_set)
        queries_by_subtree.update(subtree_query_set)

        if len(queries_by_subtree) > 0:
            subtree_dict[subtree] = (seed_label, queries_by_subtree)
            subtree_leaf_label_dict[subtree] = subtree.label_to_node(
                    selection='leaves')

        votes_b4 = len(list(lf_votes.elements()))
        for query in queries_by_subtree:
            if query in query_votes_dict:
                lf_votes.subtract(query_votes_dict[query])
                query_votes_dict.pop(query)

        if len(queries_by_subtree) == 0:
            # 10.27.2023 - Chengze Shen
            # >>> prevent going over the the total number of votes
            most_common_index += 1
        else:
            most_common_index = 0
            
    placed_query_list = []
    
    # reassign queries to the subtree minimizing total edge length 
    # from the query's top vote to the subtree's seedleaf
    new_subtree_dict = dict()
    for query, closest_label in query_top_vote_dict.items():
        best_subtree = None
        best_distance = 99999999999999999
        for subtree, value in subtree_dict.items():
            leaf_label_dict = subtree_leaf_label_dict[subtree]
            seed_label, _ = value
            if closest_label in leaf_label_dict:
                distance = subtree.distance_between(
                        leaf_label_dict[closest_label],
                        leaf_label_dict[seed_label])
                if distance < best_distance:
                   best_subtree = subtree
                   best_distance = distance
        if best_subtree in new_subtree_dict:
            new_subtree_dict[best_subtree].append(query)
        else:
            new_subtree_dict[best_subtree] = [query]

    t1 = time.perf_counter()
    _LOG.info('Time to assign queries to subtrees: {} seconds'.format(t1 - t0))
    return new_subtree_dict, placed_query_list


'''
Function to assign queries to subtrees as used in SCAMPP 
(subtrees are built using the nearest leaf as the seed sequence)
'''
def buildQuerySubtrees(query_votes_dict, query_top_vote_dict,
        tree, leaf_dict, dry_run=False):
    t0 = time.perf_counter()
    _LOG.info('(SCAMPP) Building query subtree for placement...')

    if dry_run:
        return dict(), []

    # (1) go over the query seed sequences to see if any queries use 
    # the same seed sequence (i.e. subtree)
    seed_queries = dict()
    for query, closest_leaf in query_top_vote_dict.items():
        if closest_leaf not in seed_queries:           
            seed_queries[closest_leaf] = [query]
        else:
            seed_queries[closest_leaf].append(query)

    new_subtree_dict = dict()
    # assign queries to subtrees, and remove them from the pool
    # repeat until all queries are assigned
    _total = 0
    for seed_label, queries in seed_queries.items():
        ####### additional logging for tracking progress 
        _total += 1
        if _total % 1000 == 0 or _total == len(seed_queries):
            _LOG.info(f"- Built {_total}/{len(seed_queries)} subtrees")

        node_y = leaf_dict[seed_label]
        # extract [subtreesize] leaves
        if Configs.subtreetype == "h":
            labels = query_votes_dict[queries[0]]
        elif Configs.subtreetype == "n":
            labels = utils.subtree_nodes(tree, node_y, Configs.subtreesize)
        else:
            labels = utils.subtree_nodes_with_edge_length(tree, node_y,
                Configs.subtreesize)
        subtree = tree.extract_tree_with(labels)
        new_subtree_dict[subtree] = queries

    placed_query_list = []

    t1 = time.perf_counter()
    _LOG.info('Time to assign queries to subtrees: {} seconds'.format(t1 - t0))
    return new_subtree_dict, placed_query_list

'''
Helper function to run a single placement task. Designed to use with
multiprocessing
Input: job object
Return: outpath from job.run() 
'''
def placeOneSubtree(*jobs,
        subtree_id=0, num_assigned_queries=-1, outpath=None, logging=None):
    job_type = None
    # record the last job_type and _outpath, which will be for the placement
    # job
    for job in jobs:
        job_type = job.job_type
        # run the job
        _outpath = job.run(logging=logging)
    
    # move output file for EPA-ng output
    if job_type == 'epa-ng': 
        os.system('mv {} {}'.format(_outpath, outpath))
    return subtree_id, num_assigned_queries, outpath

'''
Function to perform placement of queries for each subtree
'''
def placeQueriesToSubtrees(tree, leaf_dict, new_subtree_dict, placed_query_list,
        aln, qaln, cmdline_args, workdir, qname_map, qname_map_rev, 
        pool, lock, dry_run=False):
    t0 = time.perf_counter()
    _LOG.info("Performing placement on each subtree with {}...".format(
        Configs.placement_method))

    if dry_run:
        return dict()

    # prepare to write an aggregated results to local
    jplace = dict()
    utils.add_edge_nbrs(tree)
    jplace["tree"] = utils.newick_edge_tokens(tree)
    placements = []

    # go over the dictionary of subtrees and their assigned queries
    # perform placement using either EPA-ng or pplacer
    final_subtree_count, total_subtrees_examined = 0, 0
    futures = []
    _LOG.info("Submitting jobs for subtree placement...")
    for subtree, query_list in new_subtree_dict.items():
        total_subtrees_examined += 1

        # empty subtree, continue
        if len(query_list) == 0:
            continue

        subtree_dir = os.path.join(workdir, f'subtree_{final_subtree_count}')
        if not os.path.isdir(subtree_dir):
            os.makedirs(subtree_dir)
        
        # name all temporary output files
        tmp_tree = os.path.join(subtree_dir, f'subtree_{final_subtree_count}.tre')
        tmp_aln = os.path.join(subtree_dir, f'subtree_{final_subtree_count}_aln.fa')
        tmp_qaln = os.path.join(subtree_dir, f'subtree_{final_subtree_count}_qaln.fa')
        tmp_output = os.path.join(subtree_dir,
                'subtree_{}_{}.jplace'.format(
                    final_subtree_count, Configs.placement_method))

        # extract corresponding ref sequences and queries
        tmp_leaf_dict = subtree.label_to_node(selection='leaves')
        if '' in tmp_leaf_dict:
            del tmp_leaf_dict['']
        tmp_ref_dict = {label : aln[label] for label in tmp_leaf_dict.keys()}
        tmp_q_dict = {name : qaln[name] for name in query_list}
        write_fasta(tmp_aln, tmp_ref_dict)
        write_fasta(tmp_qaln, tmp_q_dict)

        # process the subtree before placement
        subtree.resolve_polytomies()
        subtree.suppress_unifurcations()
        subtree.write_tree_newick(tmp_tree, hide_rooted_prefix=True)

        # 1.27.2025 - Chengze Shen
        # choose the placement method to run
        jobs = []
        if Configs.placement_method == 'epa-ng':
            job = EPAngJob(path=Configs.epang_path,
                    info_path=Configs.info_path, tree_path=tmp_tree,
                    aln_path=tmp_aln, qaln_path=tmp_qaln,
                    outdir=subtree_dir, num_cpus=Configs.cpus_per_job)
            jobs.append(job)
            ## for EPA-ng, ensure that outpath name is changed to the one we want
            #_outpath = job.run(logging=f'subtree_{final_subtree_count}')
            #os.system('mv {} {}'.format(_outpath, tmp_output))
        elif Configs.placement_method == 'pplacer':
            # build ref_pkg with info and tmp_tree and tmp_aln
            refpkg_dir = os.path.join(subtree_dir,
                    f'subtree_{final_subtree_count}.refpkg')
            taxit_job = TaxtasticJob(path=Configs.taxit_path,
                    outdir=refpkg_dir, name=f'subtree_{final_subtree_count}',
                    aln_path=tmp_aln, tree_path=tmp_tree,
                    info_path=Configs.info_path)
            jobs.append(taxit_job)
            #_ = taxit_job.run()

            # run pplacer-taxtastic
            job = PplacerTaxtasticJob(path=Configs.pplacer_path,
                    refpkg_dir=refpkg_dir,
                    #molecule=Configs.molecule, model=Configs.model,
                    outpath=tmp_output, num_cpus=Configs.cpus_per_job,
                    qaln_path=tmp_qaln)
            #tmp_output = job.run(logging=f'subtree_{final_subtree_count}')
            jobs.append(job)
        else:
            raise ValueError(
                    f"Placement method {Configs.placement_method} not recognized")
        logging = f'subtree_{final_subtree_count}'
        futures.append(pool.submit(placeOneSubtree, *jobs,
            subtree_id=final_subtree_count,
            num_assigned_queries=len(query_list),
            outpath=tmp_output, logging=logging))
        # increment final_subtree_count
        final_subtree_count += 1

    # deal with outputs
    for future in concurrent.futures.as_completed(futures):
        subtree_id, num_assigned_queries, tmp_output = future.result()
        _LOG.info('- Subtree {}/{} with {} queries'.format(
            subtree_id + 1, final_subtree_count, num_assigned_queries))

        # read in each placement result
        place_file = open(tmp_output, 'r')
        place_json = json.load(place_file)
        tgt = "n"
        if Configs.placement_method == 'pplacer':
            tgt = "nm"
        if len(place_json["placements"]) > 0:
            added_tree, edge_dict = utils.read_tree_newick_edge_tokens(
                    place_json["tree"])

            # obtain the fields for "p"
            fields = place_json["fields"]
            # set the fields in jplace accordingly
            if "fields" not in jplace:
                jplace["fields"] = fields
            field_to_idx = {field: i for i, field in enumerate(fields)}

            for tmp_place in place_json["placements"]:
                # Fixed @ 7.7.2025 - Chengze Shen
                #   - pplacer actually can report multiple items
                #   - in the ["nm"] field.
                for _idx in range(len(tmp_place[tgt])):
                    # convert qname back using qname_map_rev
                    tmp_name = tmp_place[tgt][_idx]

                    # >EPA-ng: tgt=="n" --> qname is string
                    if isinstance(tmp_name, str):
                        qname = qname_map_rev[tmp_name]
                        tmp_place[tgt][_idx] = qname
                    # >pplacer: tgt=="nm" --> qname is a list of two fields
                    elif isinstance(tmp_name, list):
                        qname = qname_map_rev[tmp_name[0]]
                        tmp_place[tgt][_idx][0] = qname
                    placed_query_list.append(qname)

                #placed_query_list.append(tmp_place[tgt][0])
                for i in range(len(tmp_place["p"])):
                    edge_num = tmp_place["p"][i][
                            field_to_idx['edge_num']]
                    edge_distal = tmp_place["p"][i][
                            field_to_idx['distal_length']]

                    right_n = edge_dict[str(edge_num)]
                    left_n = right_n.get_parent()

                    # left and right path_l and path_r are in added_tree
                    left, path_l = utils.find_closest(left_n, {left_n, right_n})
                    right, path_r = utils.find_closest(right_n, {left_n, right_n})

                    left = leaf_dict[left.get_label()]
                    right = leaf_dict[right.get_label()]
                    _, path = utils.find_closest(left, {left}, y=right)
                    # now left right and path are in tree

                    length = sum([x.get_edge_length() for x in path_l])+edge_distal
                    # left path length through subtree before placement node

                    target_edge = path[-1]

                    for j in range(len(path)):
                        length -= path[j].get_edge_length()
                        if length < 0:
                            target_edge = path[j]
                            break

                    #tmp_place["p"][i][field_to_idx['edge_num']] = 0

                    label = target_edge.get_label()

                    [taxon, target_edge_nbr] = label.split('%%',1)
                    tmp_place["p"][i][field_to_idx['distal_length']] = \
                            target_edge.get_edge_length()+length
                    tmp_place["p"][i][field_to_idx['edge_num']] = \
                            int(target_edge_nbr)

                placements.append(tmp_place.copy())
        place_file.close()

    _LOG.info(f'Final number of subtrees used: {final_subtree_count}')

    # prepare the output jplace to write
    jplace["placements"] = placements
    jplace["metadata"] = {"invocation": " ".join(cmdline_args)}
    jplace["version"] = 3
    #jplace["fields"] = ["distal_length", "edge_num", "like_weight_ratio", \
    #        "likelihood", "pendant_length"]

    t1 = time.perf_counter()
    _LOG.info('Time to place queries to subtrees: {} seconds'.format(t1 - t0))
    return jplace


'''
Function to write a given jplace object to local output
'''
def writeOutputJplace(output_jplace, dry_run=False):
    t0 = time.perf_counter()
    _LOG.info('Writing aggregated placements to local...')
    
    if dry_run:
        return

    outpath = os.path.join(Configs.outdir, Configs.outname)
    outf = open(outpath, 'w')
    json.dump(output_jplace, outf, sort_keys=True, indent=4)
    outf.close()

    t1 = time.perf_counter()
    _LOG.info('Time to build final jplace file: {} seconds'.format(t1 - t0))



import os, shutil, subprocess, stat, re, traceback, shlex
from subprocess import Popen
from abc import abstractmethod

from bscampp import get_logger, log_exception
#from bscampp.configs import Configs

_LOG = get_logger(__name__)

'''
Template class Job for running external software/jobs
'''
class Job(object):
    def __init__(self):
        self.job_type = ""
        self.errors = []
        self.b_ignore_error = False
        self.pid = -1
        self.returncode = 0

    def __call__(self):
        return self.run()

    def get_pid(self):
        return self.pid

    # run the job with given invocation and raise errors when encountered
    def run(self, stdin="", lock=None, logging=None, shell=False):
        try:
            cmd, outpath = self.get_invocation()
            _LOG.debug(f'Running job_type: {self.job_type}, output: {outpath}')

            # failsafe for NotImplemented jobs
            if len(cmd) == 0:
                raise ValueError(
                        f'{self.job_type} does not have a valid run command. '
                        'It might be due to (invalid input type, etc.).')

            # identify binaries as the first field
            binpath = cmd[0]
            # deal with special cases, e.g., python, java
            if binpath == 'java':
                binpath = cmd[2]
            elif binpath == 'python' or binpath == 'python3':
                binpath = cmd[1]
            assert os.path.exists(binpath) or binpath == 'gzip', \
                    ('executable for %s does not exist: %s' %
                        (self.job_type, binpath))
            assert \
                (binpath.count('/')== 0 or os.path.exists(binpath)), \
                ('path for %s does not exist (%s)' %
                    (self.job_type, binpath))

            _LOG.debug('Arguments: %s', ' '.join(
                (str(x) if x is not None else '?NoneType?' for x in cmd)))

            # logging to local or to PIPE
            stderr, stdout = '', ''
            scmd = ' '.join(cmd)
            if logging != None:
                logpath = os.path.join(
                        os.path.dirname(outpath),
                        f'{logging}_{self.job_type}.txt')
                outlogging = open(logpath, 'w', 1)

                # TODO: may need to deal with piping in the future, for now
                # it is not needed
                p = Popen(cmd, text=True, bufsize=1,
                        stdin=subprocess.PIPE,
                        stdout=outlogging, stderr=outlogging)
                self.pid = p.pid
                stdout, stderr = p.communicate(input=stdin)
                # stdout and stderr are both written to outlogging
                # hence, assign them to be empty strings
                stdout, stderr = '', ''
                outlogging.close()
            else:
                p = Popen(cmd, text=True, bufsize=1,
                        stdin=subprocess.PIPE,
                        stdout=subprocess.PIPE, stderr=subprocess.PIPE)
                self.pid = p.pid
                stdout, stderr = p.communicate(input=stdin)
            self.returncode = p.returncode

            # successful run and write to log
            if self.returncode == 0:
                if lock:
                    try:
                        lock.acquire()
                        _LOG.debug(f'{self.job_type} completed, output: {outpath}')
                    finally:
                        lock.release()
                else:
                    _LOG.debug(f'{self.job_type} completed, output: {outpath}')
                return outpath
            else:
                error_msg = ' '.join([f'Error occurred running {self.job_type}.',
                    f'returncode: {self.returncode}'])
                if logging != None:
                    logpath = '\nLOGPATH: ' + os.path.join(
                            os.path.dirname(outpath),
                            f'{logging}_{self.job_type}.txt')
                else:
                    logpath = ''
                if lock:
                    try:
                        lock.acquire()
                        _LOG.error(error_msg + '\nSTDOUT: ' + stdout +
                                '\nSTDERR: ' + stderr + logpath)
                    finally:
                        lock.release()
                else:
                    _LOG.error(error_msg + '\nSTDOUT: ' + stdout +
                            '\nSTDERR: ' + stderr + logpath)
                exit(self.returncode)
        except Exception:
            log_exception(_LOG)

    # implemented in subclass
    # return: (cmd, outpath)
    @abstractmethod
    def get_invocation(self):
        raise NotImplementedError(
            'get_invocation() should be implemented by subclasses.')

'''
Generic job that runs the given command, represented as a list of strings
'''
class GenericJob(Job):
    def __init__(self, cmd=[], job_type='external'):
        Job.__init__(self)
        self.job_type = job_type
        self.cmd = cmd

    def get_invocation(self):
        return self.cmd, None

'''
A EPA-ng job that runs EPA-ng with given parameters
'''
class EPAngJob(Job):
    def __init__(self, **kwargs):
        Job.__init__(self)
        self.job_type = 'epa-ng'

        self.path = ''
        self.info_path = ''
        self.tree_path = ''
        self.aln_path = ''
        self.qaln_path = ''
        self.outdir = ''
        #self.molecule = ''
        #self.model = ''
        self.num_cpus = 1

        for k, v in kwargs.items():
            setattr(self, k, v)

    def get_invocation(self):
        self.outpath = os.path.join(self.outdir, 'epa_result.jplace') 
        cmd = [self.path,
                '-m', self.info_path,
                '-t', self.tree_path, '-s', self.aln_path,
                '-q', self.qaln_path, '-w', self.outdir,
                '-T', str(self.num_cpus), '--redo']
        return cmd, self.outpath

'''
A taxtastic job that create a refpkg based on given parameters
'''
class TaxtasticJob(Job):
    def __init__(self, **kwargs):
        Job.__init__(self)
        self.job_type = 'taxit'

        self.path = ''
        self.outdir = ''
        self.name = ''
        self.tree_path = ''
        self.aln_path = ''
        self.info_path = ''

        for k, v in kwargs.items():
            setattr(self, k, v)

    def get_invocation(self):
        self.outpath = os.path.join(self.outdir)
        cmd = [self.path, 'create', '-P', self.outdir,
                '-l', self.name, '--aln-fasta', self.aln_path,
                '--tree-file', self.tree_path]
        # check which model file is provided
        if 'bestModel' in self.info_path:
            cmd.extend(['--model-file', self.info_path])
        else:
            cmd.extend(['--tree-stats', self.info_path])
        return cmd, self.outpath

'''
A pplacer job that uses taxtastic refpkg to place sequences
'''
class PplacerTaxtasticJob(Job):
    def __init__(self, **kwargs):
        Job.__init__(self)
        self.job_type = 'pplacer'

        self.path = ''
        self.refpkg_dir = ''
        self.qaln_path = ''
        self.outdir = ''
        self.outpath = ''
        #self.model = 'GTR'
        self.num_cpus = 1

        for k, v in kwargs.items():
            setattr(self, k, v)

    def get_invocation(self):
        # outpath defined 
        cmd = [self.path,
                #'-m', self.model,
                '-c', self.refpkg_dir, '-o', self.outpath,
                '-j', str(self.num_cpus), self.qaln_path]
        return cmd, self.outpath



############################################################
#
#   Init file for BSCAMPP, using the __init__.py from
#   SEPP as the original template. Current adaption comes
#   from https://github.com/c5shen/TIPP3.git
#
############################################################
from operator import itemgetter
import logging, os

# update system recursion limit to avoid issues
# not really needed for BSCAMPP but safe to update here
os.sys.setrecursionlimit(1000000)

__version__ = "1.0.8"
_INSTALL_PATH = __path__[0]

# global variables to store all loggers
__set_loggers = set()

# obtain the current logging level, default to INFO
def get_logging_level(logging_level='info'):
    logging_level_map = {
            'DEBUG': logging.DEBUG, 'INFO': logging.INFO,
            'WARNING': logging.WARNING, 'ERROR': logging.ERROR,
            'CRITICAL': logging.CRITICAL,
            }
    # obtain from environment variable to determine logging level, if
    # set by the user
    env_level = os.getenv('BSCAMPP_LOGGING_LEVEL')
    if env_level is not None:
        ll = env_level.upper()
    else:
        ll = logging_level.upper()
    # default to INFO if ll is not defined
    return logging_level_map.get(ll, logging.INFO)

# obtain a logger for a given file
def get_logger(name='bscampp', log_path=None, logging_level='info'):
    logger = logging.getLogger(name)
    if name not in __set_loggers:
        # set up a new logger for a name not in __set_loggers yet
        level = get_logging_level(logging_level)
        logging_formatter = logging.Formatter(
            ("[%(asctime)s] %(filename)s (line %(lineno)d):"
             " %(levelname) 8s: %(message)s"))
        logging_formatter.datefmt = "%H:%M:%S"
        logger.setLevel(level)

        # logging to stdout
        if log_path is None:
            ch = logging.StreamHandler()
        else:
            # use FileHandler for logging
            ch = logging.FileHandler(log_path, mode='a')
        ch.setLevel(level)
        ch.setFormatter(logging_formatter)
        logger.addHandler(ch)
        __set_loggers.add(name)
    return logger

# logging exception
def log_exception(logger):
    import traceback, io
    s = io.StringIO()
    traceback.print_exc(None, s)
    logger.error(s.getvalue())
    exit(1)



import os, time
try:
    import configparser
except ImportError:
    from ConfigParser import configparser
from argparse import ArgumentParser, Namespace
import subprocess

from bscampp.init_configs import init_config_file
from bscampp import get_logger, log_exception
#from bscampp.utils import inferDataType

# detect home.path or create if missing
homepath = os.path.dirname(__file__) + '/home.path'
_root_dir, main_config_path = init_config_file(homepath)

# set valid configparse section names
valid_config_sections = []
logging_levels = set(['DEBUG', 'INFO', 'WARNING', 'ERROR', 'CRITICAL'])

_LOG = get_logger(__name__)

'''
Configuration defined by users and by default values
'''
class Configs:

    # basic input paths
    info_path = None        # info file for pplacer or EPA-ng
    tree_path = None        # placement tree path
    aln_path = None         # alignment for backbone. Optinally with queries
    qaln_path = None        # (optional) alignment for query.
    outdir = None           # output directory
    outname = None          # output name for the final jplace file
    keeptemp = False        # whether to keep all temporary files
    verbose = 'INFO'        # default verbose level to print
    num_cpus = 1            # number of cores to use for parallelization
    cpus_per_job = 2        # number of cores to use per job
    max_workers = 1         # max_workers for ProcessPoolExecutor
                            # ... = max(1, num_cpus // cpus_per_job)

    # binaries
    pplacer_path = None
    epang_path = None
    taxit_path = None
    hamming_distance_dir = None

    # placement settings
    placement_method = 'epa-ng'
    subtreesize = 2000
    votes = 5
    similarityflag = True

    # miscellaneous
    tmpfilenbr = 0
    fragmentflag = True
    subtreetype = 'd'

# check if the given configuration is valid to add
def set_valid_configuration(name, conf):
    if not isinstance(conf, Namespace):
        _LOG.warning(
            "Looking for Namespace object from \'{}\' but find {}".format(
                name, type(conf)))
        return

    # basic section defined in main.config
    if name == 'basic':
        for k in conf.__dict__.keys():
            k_attr = getattr(conf, k)
            if not k_attr:
                continue
            if k in Configs.__dict__:
                setattr(Configs, k, k_attr)
    else:
        pass

# valid attribute check for print out
def valid_attribute(k, v):
    if not isinstance(k, str):
        return False
    if k.startswith('_'):
        return False
    return True

# print out current configuration
def getConfigs():
    msg = '\n************ Configurations ************\n' + \
            f'\thome.path: {homepath}\n' + \
            f'\tmain.config: {main_config_path}\n\n'
    for k, v in Configs.__dict__.items():
        if valid_attribute(k, v):
            msg += f'\tConfigs.{k}: {v}\n'
    print(msg, flush=True)

# read in config file if it exists
def _read_config_file(filename, cparser, opts,
        child_process=False, expand=None):
    config_defaults = []
    with open(filename, 'r') as f:
        cparser.read_file(f)
        if cparser.has_section('commandline'):
            for k, v in cparser.items('commandline'):
                config_defaults.append(f'--{k}')
                config_defaults.append(v)

        for section in cparser.sections():
            if section == 'commandline':
                continue
            if getattr(opts, section, None):
                section_name_space = getattr(opts, section)
            else:
                section_name_space = Namespace()
            for k, v in cparser.items(section):
                if expand and k == 'path':
                    v = os.path.join(expand, v)
                setattr(section_name_space, k, v)
            setattr(opts, section, section_name_space)
    return config_defaults

# validate a given binary file is executable
def validate_binary_executable(name, path):
    # 1. make sure path exists
    if not os.path.exists(path):
        raise FileNotFoundError(f"{path} is not found!")

    # 2. make sure path is executable with no IO error
    try:
        p = subprocess.Popen([path], text=True, stdout=subprocess.PIPE,
                stderr=subprocess.PIPE)
        _, __ = p.communicate()
    except OSError:
        raise AssertionError(' '.join([
            f"{path} is not executable! Please do the followings:",
            f"\n\n\t1. obtain a working executable for the software: {name}",
            f"\n\t2. modify {main_config_path} to replace the old executable path\n",
            ]))

'''
Build Config class
'''
def buildConfigs(parser, cmdline_args, child_process=False, rerun=False):
    cparser = configparser.ConfigParser()
    cparser.optionxform = str
    args = parser.parse_args(cmdline_args)

    # Check if only updating config files, if so, re-initialize the 
    # configuration file at ~/.bscampp/main.config and exit
    #if args.command == 'update-configs':
    #    _ = init_config_file(homepath, rerun=True)
    #    _LOG.warning('Finished re-initializing the configuration file '
    #            f'at {main_config_path}, exiting...')
    #    exit(0)

    # first load arguments from main.configs
    main_args = Namespace()
    cmdline_main = _read_config_file(main_config_path,
            cparser, main_args, child_process=child_process)

    # merge arguments, in the correct order so things are overridden correctly
    args = parser.parse_args(cmdline_main + cmdline_args,
            namespace=main_args)

    # directly add all arguments that's defined in the Configs class
    for k in args.__dict__.keys():
        k_attr = getattr(args, k)
        if k in Configs.__dict__:
            # valid argument that's defined in the Configs class
            setattr(Configs, k, k_attr)
        else:
            # check if the argument is valid
            set_valid_configuration(k, k_attr)

    # create outdir
    if not os.path.isdir(Configs.outdir):
        os.makedirs(Configs.outdir)

    # modify outname if it does not have a .jplace suffix
    if Configs.outname.split('.')[-1].lower() != 'jplace':
        Configs.outname += '.jplace'

    # modify num_cpus if it is the default value
    if Configs.num_cpus > 0:
        Configs.num_cpus = min(os.cpu_count(), Configs.num_cpus)
    else:
        Configs.num_cpus = os.cpu_count()
    # compute max_workers based on num_cpus and cpus_per_job
    Configs.max_workers = max(1, Configs.num_cpus // Configs.cpus_per_job)

    # sanity check for existence of base placement binary path
    if Configs.placement_method == 'epa-ng':
        validate_binary_executable(Configs.placement_method, Configs.epang_path)
        #assert os.path.exists(Configs.epang_path), 'epa-ng not detected!'
    elif Configs.placement_method == 'pplacer':
        validate_binary_executable(Configs.placement_method, Configs.pplacer_path)
        #assert os.path.exists(Configs.pplacer_path), 'pplacer not detected!'



#from dendropy import *
import numpy as np
import heapq
import treeswift
import itertools
from collections import deque
from os.path import expanduser,isfile
import random
import statistics
import copy
import gzip
import argparse

from bscampp import get_logger, log_exception
_LOG = get_logger(__name__)

# reformat argparse help text formatting
class SmartHelpFormatter(argparse.RawDescriptionHelpFormatter):
    def add_text(self, text):
        if text is not None:
            text = text.replace("\\n", "\n").replace("\\t", "\t")
        super().add_text(text)
    def _split_lines(self, text, width):
        if '\n' in text:
            temp = text.split('\n')
            ret = []
            for _splice in [argparse.RawDescriptionHelpFormatter._split_lines(self, x, width)
                    for x in temp]:
                ret.extend(_splice)
            return ret
        return argparse.RawDescriptionHelpFormatter._split_lines(self, text, width)

# store bracket open/close for convenience in label parsing
BRACKET = {
    '[': ']', # square bracket
    '{': '}', # curly bracket
    "'": "'", # single-quote
    '"': '"', # double-quote
}


# infer datatype from input file
def inferDataType(path):
    sequences = read_data(path)
    acg, t, u, total = 0, 0, 0, 0
    for taxon, seq in sequences.items():
        letters = seq.upper()
        for letter in letters:
            total = total + 1

            if letter in ('A', 'C', 'G', 'N'):
                acg += 1
            elif letter == 'T':
                t += 1
            elif letter == 'U':
                u += 1
    # dna -> nucleotide
    if u == 0 and (acg + t) / total > 0.9:
        datatype = 'nucleotide'
    # rna -> nucleotide
    elif t == 0 and (acg + u) / total > 0.9:
        datatype = 'nucleotide'
    # amino acid -> protein
    else:
        datatype = 'protein'

    _LOG.info(f"Inferred input data type: {datatype}")
    return datatype

def write_fasta(aln, aln_dict, aligned=True):
    """ Write given dictionary as FASTA file out
    
    Parameters
    ----------
    aln : FASTA file path
    aln_dict : MSA in the form of a dict
    aligned : whether the sequences are aligned

    Returns
    -------
    None

    """

    f = open(aln, 'w')
    for label, seq in aln_dict.items():
        if label != '':
            f.write(f'>{label}\n')
            if aligned:
                f.write(f'{seq}\n')
            else:
                f.write(seq.replace('-', '') + '\n')
    f.close()

#separete the query and ref sequence from the alignment file

def read_data(aln):
    """ Load the query and reference sequence from the alignment file
    
    Parameters
    ----------
    aln : multiple sequence alignment containing reference taxa and query sequence
    
    Returns
    -------
    dictionary containing sequences with taxon label keys
    
    """

    # determine the file type, whether we have a .gz/.gzip file
    suffix = aln.split('.')[-1]
    if suffix in ['gz', 'gzip']:
        f = gzip.open(aln, 'rt')
    else:
        f = open(aln)
    result = dict()

    taxa = ""
    seq = ""
    for line in f:
        if line[0] == '>':
            if taxa != "":
                result[taxa] = seq
            taxa = line[1:-1]
            seq = ""

        elif line == "/n":
            continue
        else:
            seq += line[:-1]

    if taxa != "":
        result[taxa] = seq


    return result

def seperate(aln_dict, leaf_dict):
    """ Separate the query sequences from the reference sequences
    
    Parameters
    ----------
    aln_dict : Sequence dictionary with taxon label keys
    leaf_dict : Sequence dictionary with leaf label keys (queries are not in backbone tree)
    
    Returns
    -------
    separate dictionaries containing query sequences and referece sequences with taxon label keys
    
    """
    ref = dict()
    query = dict()

    for key, value in aln_dict.items():
        if key not in leaf_dict:
            query[key] = value
        else:
            ref[key] = value

    return ref, query

def hamming(seq1, seq2):
    """ Returns hamming distance between two sequences
    
    Parameters
    ----------
    seq1 : query sequence
    seq2 : reference sequence
    
    Returns
    -------
    integer hamming distance between query sequence and reference sequence
    
    """
    return sum(1 for ch1, ch2 in zip(seq1, seq2) if ch1 != ch2)


def find_y(x,ref):
    """ Returns leaf label for closest sister taxon l (no longer used)
    
    Parameters
    ----------
    x : aligned query sequence
    ref : reference multiple sequence alignment dictionary 
    
    Returns
    -------
    leaf label for taxon with smallest hamming distacne to query sequence
    
    """
    low = len(x)
    y = ""
    for name, seq in ref.items():
        h_dist = hamming(x, seq)
        if h_dist < low:
            low = h_dist
            y = name
    return y

def set_hamming_dict(args):
    dict_items, other_args = args
    name, seq = dict_items
    ref, n, fragment_flag, y_dict = other_args
    
    y_dict[name] = find_closest_hamming(seq, ref, n, fragment_flag)


def find_closest_hamming(x, ref, n, fragment_flag):
    ''' Returns leaf name for n closest sister taxa to sequence x
    
    Parameters
    ----------
    x : aligned query sequence
    ref : reference multiple sequence alignment dictionary 
    n : number of nodes to return 
    fragment_flag : True if the query is not full length
    
    Returns
    -------
    list of nodes with n smallest hamming distacnes to query sequence

    '''
    queue = []
    closest = []

    counter = 0
    if fragment_flag == 'True':
        [si, ei] = set_fragment_indicies(x)
    else:
        [si, ei] = [0, len(x)]
    
    c = 200 # size of the subtring compared at once
        
    for name, seq in ref.items():
        heapq.heappush(queue,(hamming(seq[si:si+c],x[si:si+c]), ei - si - c, counter, name))
        counter += 1

    while queue:
        (ham_dist, sites_left, cnt, name) = heapq.heappop(queue)
        if sites_left < 0:
            closest.append(name)
            if len(closest) >= n:
                return closest
        else:
            ind = ei - sites_left
            new_ham = hamming(ref[name][ind:ind+c],x[ind:ind+c])
            heapq.heappush(queue,(ham_dist + new_ham, sites_left - c, cnt, name))

def set_fragment_indicies(x):
    """ Returns the indicees without leading and trailing gaps.
    
    Parameters
    ----------
    x = string sequence
    
    Returns
    -------
    list of start index and end index with the first and last non gap character
    
    """
    e = len(x)
    ei = e
    si = 0
    for i in range(ei): 
        if x[i] == '-' and si == i:
            si = i + 1
        if x[e - i - 1] == '-' and ei == e - i:
            ei = e - i - 1
        if ei == si:
            break
    return [si, ei] 

def find_closest(x, visited, y=None):
    """ Returns leaf label for closest leaf to the node x through path not travelling through visited.
    If y is populated returns path from x to y not travelling through nodes in visited.
    
    Parameters
    ----------
    x : dendropy node object
    visited : dictionary containing dendropy node objects as keys 
    y : dendropy node object
    
    Returns
    -------
    If y == None : dendropy node object of closest leaf y to the node x through path not travelling through nodes in visited, 
                   list containing dendropy node objects on path to that leaf y from node x
    If y != None : dendropy node object y, 
                   list containing dendropy node objects on path from node x to leaf y not travelling through nodes in visited
    
    """
    queue = []
    cnt = 1
    visited.add(x)

    if x.get_parent() and x.get_parent() not in visited:
        tmp = []
        tmp.append(x)
        heapq.heappush(queue, [x.get_edge_length(), cnt, tmp, x.get_parent()])
        cnt += 1

    for child in x.child_nodes():
        if child and child not in visited:
            tmp = []
            tmp.append(child)
            heapq.heappush(queue, [child.get_edge_length(), cnt, tmp, child])
            cnt += 1

    while len(queue) > 0:
        try:
            [length, _, path, node] = heapq.heappop(queue)
        except IndexError:
            break

        visited.add(node)
        if node.is_leaf():
            if (not y) or node.get_label()==y.get_label():
                return node, path
            else:
                continue

        if node.get_parent() and node.get_parent() not in visited:
            tmp = path.copy()
            tmp.append(node)
            heapq.heappush(queue, [length+node.get_edge_length(), cnt, tmp, node.get_parent()])
            cnt += 1

        for child in node.child_nodes():
            if child and child not in visited:
                tmp = path.copy()
                tmp.append(child)
                heapq.heappush(queue, [length+child.get_edge_length(), cnt, tmp, child])
                cnt += 1

    return x, [x]


def find_closest_testing(x, visited, y=None, valid_leaves=None):
    """ Returns leaf label for closest leaf to the node x through path not travelling through visited.
    If y is populated returns path from x to y not travelling through nodes in visited.
    
    Parameters
    ----------
    x : dendropy node object
    visited : dictionary containing dendropy node objects as keys
    y : dendropy node object
    
    Returns
    -------
    If y == None : dendropy node object of closest leaf y to the node x through path not travelling through nodes in visited, 
                   list containing dendropy node objects on path to that leaf y from node x
                   distance from node x to leaf y
    If y != None : dendropy node object y, 
                   list containing dendropy node objects on path from node x to leaf y not travelling through nodes in visited, 
                   distance from node x to leaf y
    
    """
    queue = []
    cnt = 1
    visited.add(x)

    if x.get_parent() and x.get_parent() not in visited:
        tmp = []
        tmp.append(x)
        heapq.heappush(queue, [x.get_edge_length(), cnt, tmp, x.get_parent()])
        cnt += 1

    for child in x.child_nodes():
        if child and child not in visited:
            tmp = []
            tmp.append(child)
            heapq.heappush(queue, [child.get_edge_length(), cnt, tmp, child])
            cnt += 1

    while len(queue) > 0:
        try:
            [length, _, path, node] = heapq.heappop(queue)
        except IndexError:
            break

        visited.add(node)
        if node.is_leaf():
            if (not y) or node.get_label() == y.get_label():
                if  (valid_leaves != None and node.get_label() not in valid_leaves):
                    continue
                else:
                    return node, path, length
            else:
                continue

        if node.get_parent() and node.get_parent() not in visited:
            tmp = path.copy()
            tmp.append(node)
            heapq.heappush(queue, [length+node.get_edge_length(), cnt, tmp, node.get_parent()])
            cnt += 1

        for child in node.child_nodes():
            if child and child not in visited:
                tmp = path.copy()
                tmp.append(child)
                heapq.heappush(queue, [length+child.get_edge_length(), cnt, tmp, child])
                cnt += 1

    return x, [x], 0

def build_subtrees(sister_taxon_dict, leaf_dict, tree, nbr_subtrees, subtree_size):
    tree_building_taxa = len(list(sister_taxon_dict))
    cluster_dict = dict()
    cluster_index = 0 
    cluster_list = dict()
    
    

    #tmp
    #draw(tree.extract_tree_with(list(sister_taxon_dict)))
    #tmp

    for taxon, _ in sister_taxon_dict.items():
        cluster_dict[taxon] = cluster_index
        cluster_list[cluster_index] = [taxon]
        cluster_index += 1

    queue = []
    for l1, _ in sister_taxon_dict.items():
        node_l2, _, length = find_closest_testing(leaf_dict[l1],{leaf_dict[l1]},valid_leaves=sister_taxon_dict)
        heapq.heappush(queue, [length, l1, node_l2.get_label()])

    while tree_building_taxa > nbr_subtrees:        
        [length, l1, l2] = heapq.heappop(queue)

        idx = cluster_dict[l1]
        old_idx = cluster_dict[l2]

        if idx == old_idx:
            continue

        for leaf in cluster_list[old_idx]:
            cluster_dict[leaf] = idx
        #[cluster_dict[leaf] = idx for leaf in cluster_list[old_idx]]
        cluster_list[idx].extend(cluster_list[old_idx])
        cluster_list.pop(old_idx)

        tree_building_taxa -= 1

        best_leaf = cluster_list[idx][0]
        for leaf in cluster_list[idx]:
            best_leaf = centered_leaf(tree, cluster_list[idx])
        cluster_node_list = [leaf_dict[leaf] for leaf in cluster_list[idx]]
        cluster_set = {*cluster_node_list}
        node_l3, _, length = find_closest_testing(leaf_dict[best_leaf], cluster_set, valid_leaves=sister_taxon_dict)
        heapq.heappush(queue, [length, best_leaf, node_l3.get_label()])

    trees = []
    query_decomp_dict = []

    for idx, cluster in cluster_list.items():
        best_leaf = centered_leaf(tree, cluster)

        subtree_leaves, tree2 = subtree_with_edge_length(tree, leaf_dict[best_leaf], subtree_size)
        tree2 = tree2.extract_tree_with(subtree_leaves)
        trees.append(tree2)
        query_list = []
        for leaf in cluster:
            query_list.extend(sister_taxon_dict[leaf])
        query_decomp_dict.append(query_list)

    return trees, query_decomp_dict


def subtree_nodes(tree, leaf_y, n):
    """ Returns list of length n of leaves closest to sister taxon
    
    Parameters
    ----------
    tree : treeswift tree object
    leaf_y : treeswift node for closest sister taxon
    n = number of taxa contained in subtree
    
    Returns
    -------
    list of taxon labels corresponding to leaves in the subtree
    
    """
    queue = [(0, 0, leaf_y.get_parent())]

    leaves = [leaf_y]
    visited = {leaf_y}

    counter = 1

    while len(leaves) < n:
        try:
            (length, _, node) = heapq.heappop(queue)
        except IndexError:
            break

        visited.add(node)
        if node.is_leaf():
            leaves.append(node)

        adjacent = node.child_nodes()
        if not node.is_root():
            adjacent.append(node.get_parent())

        for neighbor in adjacent:
            if neighbor not in visited:
                heapq.heappush(queue, (length+1, counter, neighbor))
                counter += 1


    result = []
    for item in leaves:
        result.append(item.get_label())

    return result

def subtree_nodes_with_edge_length(tree, leaf_y, n):
    """ Returns list of length n of leaves closest to sister taxon (minimizing edge weights)
    
    Parameters
    ----------
    tree : treeswift tree object
    leaf_y : treeswift node for closest sister taxon
    n = number of taxa contained in subtree
    
    Returns
    -------
    list of taxon labels corresponding to leaves in the subtree
    """
    queue = [(leaf_y.get_edge_length(), leaf_y.get_parent())]

    leaves = [leaf_y]
    visited = {leaf_y}

    while len(leaves) < n:
        try:
            (length, node) = heapq.heappop(queue)
        except IndexError:
            break

        visited.add(node)
        if node.is_leaf() and node.get_label() != '':
            leaves.append(node)

        adjacent_nodes = node.child_nodes()
        if not node.is_root():
            adjacent_nodes.append(node.get_parent())

        for neighbor in adjacent_nodes:
            if neighbor not in visited:
                if neighbor == node.get_parent():
                    heapq.heappush(queue, (length+node.get_edge_length(), neighbor))
                else:
                    heapq.heappush(queue, (length+neighbor.get_edge_length(), neighbor))

    result = []
    for item in leaves:
        result.append(item.get_label())

    return result
    
def subtree_with_edge_length(tree, leaf_y, n):
    """ Returns a subtree and list of length n of leaves closest to sister taxon (minimizing edge weights)
    
    Parameters
    ----------
    tree : treeswift tree object
    leaf_y : treeswift node for closest sister taxon
    n = number of taxa contained in subtree
    
    Returns
    -------
    a subtree and a list of taxon labels corresponding to leaves in the subtree
    """
    if leaf_y.get_parent() == None:
        return None, None
    
    subtree = treeswift.Tree(); subtree.root.label = None; subtree.root.edge_length = 0
    subtreeNode = treeswift.Node(label=leaf_y.get_label(), edge_length=0); subtree.root.add_child(subtreeNode)
        
    queue = [(leaf_y.get_edge_length(), leaf_y.get_parent(), leaf_y.get_edge_length(), subtree.root)]

    leaves = [leaf_y]
    visited = {leaf_y}

    while len(leaves) < n:
        try:
            (length, node, n_edge_length, parent) = heapq.heappop(queue)
        except IndexError:
            break

        visited.add(node)
        subtreeNode = treeswift.Node(label=node.get_label(), edge_length=n_edge_length); parent.add_child(subtreeNode)
        if node.is_leaf():
            leaves.append(node)

        adjacent_nodes = node.child_nodes()
        if not node.is_root():
            adjacent_nodes.append(node.get_parent())

        for neighbor in adjacent_nodes:
            if neighbor not in visited:
                if neighbor == node.get_parent():
                    heapq.heappush(queue, (length+node.get_edge_length(), neighbor, node.get_edge_length(), subtreeNode))
                else:
                    heapq.heappush(queue, (length+neighbor.get_edge_length(), neighbor, neighbor.get_edge_length(), subtreeNode))

    result = []
    for item in leaves:
        result.append(item.get_label())
    
    #subtree.deroot()
    subtree.suppress_unifurcations()
    return result, subtree
    
def extract_taxa_from_tree(a_tree, labels, leaf_dict):
    queue = []
    closest = []

    counter = 0
        
    for label in labels: 
        heapq.heappush(queue,(counter,leaf_dict[label]))
        counter += 1

    while queue:
        (cnt, node) = heapq.heappop(queue)
        if node.num_children() == 0:
            parent = node.get_parent()
            if parent != None:
                parent.remove_child(node)
                heapq.heappush(queue,(counter, parent))
    
    a_tree.resolve_polytomies()            
    a_tree.suppress_unifurcations()

    return a_tree
    
def min_tree_extract_disjoint(a_tree, max_size, tmp_leaves):
    labels, t2 = subtree_with_edge_length(a_tree, tmp_leaves[random.choice(list(tmp_leaves))], max_size)
    
    t2 = t2.extract_tree_with(labels)
    t1 = extract_taxa_from_tree(a_tree, labels, tmp_leaves)
    for label in labels:
       del tmp_leaves[label]
    return t1, t2, tmp_leaves

    
def min_tree_extract_non_disjoint(a_tree, max_size, tmp_leaves):
    labels, t2 = subtree_with_edge_length(a_tree, tmp_leaves[random.choice(list(tmp_leaves))], max_size)
    
    t2 = a_tree.extract_tree_with(labels)
    for label in labels:
       if label in tmp_leaves:
           del tmp_leaves[label]
    return a_tree, t2, tmp_leaves
    
def decompose_tree(a_tree,max_size):
    tree_list = []
    tmp_leaves = a_tree.label_to_node(selection='leaves')
    if '' in tmp_leaves:
        del tmp_leaves['']
    #t1, t2, tmp_leaves = min_tree_extract_disjoint(a_tree, max_size, tmp_leaves)
    t1, t2, tmp_leaves = min_tree_extract_non_disjoint(a_tree, max_size, tmp_leaves)
    while len(tmp_leaves) > 0:
        tree_list.append(t2)

        #t1, t2, tmp_leaves = min_tree_extract_disjoint(t1, max_size, tmp_leaves)
        t1, t2, tmp_leaves = min_tree_extract_non_disjoint(t1, max_size, tmp_leaves)
    tree_list.append(t2)
    print ("nbr subtrees: ", len(tree_list))
    return tree_list
    
def decompose_tree_min_clust(a_tree,max_size):
    tree_list = []
    t1, t2 = min_cluster_size_bisect(a_tree,max_size)
    while t2 != None:
        tree_list.append(t2)
        t1, t2 = min_cluster_size_bisect(t1, max_size)
    tree_list.append(t1) #this might be a bug
    return tree_list    
    
def min_cluster_size_bisect(a_tree,max_size):
    '''
    modified from PASTA to use treeswift
    '''
    nleaf = dict()

    print("before extracting subtree: " + str(len(a_tree.label_to_node(selection='leaves'))))
    #a_tree.draw()
    
    for node in a_tree.traverse_postorder():
        if node.is_leaf():
            nleaf[node] = 1
        else:
            nleaf[node] = 0
            max_child = None
            max_nleaf = 0
            for ch in node.child_nodes():
               nleaf[node] += nleaf[ch]
               if nleaf[ch] > max_nleaf:
                   max_nleaf = nleaf[ch]
                   max_child = ch
        if nleaf[node] >= max_size:
            node.remove_child(max_child)
            t1 = a_tree.extract_subtree(max_child)
            print("subtree size:              " + str(len(t1.label_to_node(selection='leaves'))))
            print("after extracting subtree:  " + str(len(a_tree.label_to_node(selection='leaves'))))
            #t1.deroot()
            #t1.draw()
            t1.resolve_polytomies()
            return a_tree,t1
    
    print("after extracting subtree:  " + str(len(a_tree.label_to_node(selection='leaves'))))
    return a_tree,None      

def centered_leaf(tree, cluster):
    #pick best sister leaf for cluster center
    best_leaf = cluster[0]
    if len(cluster) > 1:
        tmp_tree = tree.extract_tree_with(cluster)
        min_distance = 99999999999
        tmp_label_dict = tmp_tree.label_to_node(selection='leaves')
        for leaf in cluster: 
            total = total_distance(tmp_tree, tmp_label_dict[leaf])

            if total < min_distance:
                min_distance = total
                best_leaf = leaf

    return best_leaf

def max_distance(tree, node):
    maximum = 0 
    for leaf in tree.traverse_leaves():
        distance = tree.distance_between(node, leaf)
        if distance >= maximum:
            maximum = distance
    return maximum

def total_distance(tree, node):
    total = 0 
    for leaf in tree.traverse_leaves():
        total += tree.distance_between(node, leaf)
    return total


def avg_distance(tree, node):
    total = 0 
    leaf_cnt = 0
    for leaf in tree.traverse_leaves():
        total += tree.distance_between(node, leaf)
        leaf_cnt += 1
    return total/leaf_cnt

def median_distance(tree, node):
    total = 0
    leaf_cnt = 0
    distances = []
    for leaf in tree.traverse_leaves():
        distance = tree.distance_between(node, leaf)
        total += distance
        distances.append(distance)
        leaf_cnt += 1
    return statistics.median(distances)

def add_edge_nbrs(tree):
    counter = 0
    for node in tree.traverse_postorder():
        #if not node.is_root():
            counter += 1
            label = node.get_label()
            if label == None:
                node.set_label('%%{}'.format(counter))
            else:
                node.set_label('{}%%{}'.format(label, counter))

def remove_edge_nbrs(tree):
    for node in tree.traverse_postorder():
        #if not node.is_root():
            label_list = node.get_label().split('%%',1)
            if label_list[0] == '':
                node.set_label(None)
            else:
                node.set_label(label_list[0])

'''
   The following three functions are modified from treeswift to
   read and write newick files with jplace tokens
'''
def newick_edge_tokens(tree):
    '''
    Modified from treeswift tree.newick()
    Output this ``Tree`` as a Newick string with lables
    Returns:
        ``str``: Newick string of this ``Tree``
    '''
    label_list = tree.root.get_label().split('%%',1)

    if tree.root.edge_length is None:
        suffix = ';'
    elif isinstance(tree.root.edge_length,int):
        suffix = '%s:%d{%d};' % (str(label_list[0]), int(tree.root.edge_length), int(label_list[1]))
    elif isinstance(tree.root.edge_length,float) and tree.root.edge_length.is_integer():
        suffix = '%s:%d{%d};' % (str(label_list[0]), float(tree.root.edge_length), int(label_list[1]))
    else:
        suffix = '%s:%s{%d};' % (str(label_list[0]), str(tree.root.edge_length), int(label_list[1]))

    if tree.is_rooted:
        return '[&R] %s%s' % (newick_edge_tokens_node(tree.root),suffix)
    else:
        return '%s%s' % (newick_edge_tokens_node(tree.root),suffix)

def newick_edge_tokens_node(node):
    '''
    Modified from treeswift node.newick()
    Newick string conversion starting at this ``Node`` object
    Returns:
        ``str``: Newick string conversion starting at this ``Node`` object
    '''
    node_to_str = dict()
    for node in node.traverse_postorder():
        node_label = node.get_label()
        [label, edge_nbr] = node_label.split('%%',1)
        #node.set_label(label_list[0])
        if node.is_leaf():
            if label is None:
                node_to_str[node] = ''
            else:
                node_to_str[node] = str(label)
        else:
            out = ['(']
            for c in node.children:
                c_label = c.get_label()
                [label_c, edge_nbr_c] = c_label.split('%%',1)
                out.append(node_to_str[c])
                if c.edge_length is not None:
                    if isinstance(c.edge_length,int):
                        l_str = str(c.edge_length)
                    elif isinstance(c.edge_length,float) and c.edge_length.is_integer():
                        l_str = str(int(c.edge_length))
                    else:
                        l_str = str(c.edge_length)
                    out.append(':%s{%d}' % (l_str, int(edge_nbr_c)))
                out.append(',')
                del node_to_str[c]
            out.pop() # trailing comma
            out.append(')')
            if label is not None:
                out.append(str(label))
            node_to_str[node] = ''.join(out)
    return node_to_str[node]

#def write_tree_newick_edge_tokens(tree, filename, hide_rooted_prefix=False):
#    '''
#    Modified from treeswift tree.write_tree_newick()
#    Write this ``Tree`` to a Newick file
#       Args:
#        ``filename`` (``str``): Path to desired output file (plain-text or gzipped)
#    '''
#    if not isinstance(filename, str):
#        raise TypeError("filename must be a str")
#    treestr = newick_edge_nbr_string(tree)
#    if hide_rooted_prefix:
#        if treestr.startswith('[&R]'):
#            treestr = treestr[4:].strip()
#    else:
#        warn("Specified hide_rooted_prefix, but tree was not rooted")
#    if filename.lower().endswith('.gz'): # gzipped file
#        f = gopen(expanduser(filename),'wb',9); f.write(treestr.encode()); f.close()
#    else: # plain-text file
#        f = open(expanduser(filename),'w'); f.write(treestr); f.close()

def read_tree_newick_edge_tokens(newick):
    '''
    Modified from treeswift.read_tree_newick(newick)
    Read a tree from a Newick string or file
    Args:
        ``newick`` (``str``): Either a Newick string or the path to a Newick file (plain-text or gzipped)

    Returns:
        ``Tree``: The tree represented by ``newick``. If the Newick file has multiple trees (one per line), a ``list`` of ``Tree`` objects will be returned
    '''
    place_edge_dict = dict()
    if not isinstance(newick, str):
        try:
            newick = str(newick)
        except:
            raise TypeError("newick must be a str")
    if newick.lower().endswith('.gz'): # gzipped file
        f = gzip.open(expanduser(newick)); ts = f.read().decode().strip(); f.close()
    elif isfile(expanduser(newick)): # plain-text file
        f = open(expanduser(newick)); ts = f.read().strip(); f.close()
    else:
        ts = newick.strip()
    lines = ts.splitlines()
    if len(lines) != 1:
        return [read_tree_newick_edge_tokens(l) for l in lines]
    try:
        t = treeswift.Tree(); t.is_rooted = ts.startswith('[&R]')
        if ts[0] == '[':
            ts = ']'.join(ts.split(']')[1:]).strip(); ts = ts.replace(', ',',')
        n = t.root; i = 0
        while i < len(ts):
            # end of Newick string
            if ts[i] == ';':
                if i != len(ts)-1 or n != t.root:
                    raise RuntimeError("INVALID NEWICK")

            # go to new child
            elif ts[i] == '(':
                c = treeswift.Node(); n.add_child(c); n = c

            # go to parent
            elif ts[i] == ')':
                n = n.parent

            # go to new sibling
            elif ts[i] == ',':
                n = n.parent; c = treeswift.Node(); n.add_child(c); n = c

            # edge length
            elif ts[i] == ':':
                i += 1; ls = ''
                while ts[i] != ',' and ts[i] != ')' and ts[i] != ';' and ts[i] != '{':
                    ls += ts[i]; i += 1
                if ls[0] == '[':
                    n.edge_params = ']'.join(ls.split(']')[:-1]); ls = ls.split(']')[-1]
                n.edge_length = float(ls); i -= 1

            # edge token
            elif ts[i] == '{':
                i += 1; ls = ''
                while ts[i] != '}':
                    ls += ts[i]; i += 1
                place_edge_dict[ls] = n

            # node label
            else:
                label = ''; bracket = None
                while bracket is not None or ts[i] in BRACKET or (ts[i] != ':' and ts[i] != ',' and ts[i] != ';' and ts[i] != ')'):
                    if ts[i] in BRACKET and bracket is None:
                        bracket = ts[i]
                    elif bracket is not None and ts[i] == BRACKET[bracket]:
                        bracket = None
                    label += ts[i]; i += 1
                i -= 1; n.label = label
            i += 1
    except Exception as e:
        raise RuntimeError("Failed to parse string as Newick")
    return t, place_edge_dict



# tests/test_dry_run.py
import pytest, os
from bscampp.pipeline import bscampp_pipeline, scampp_pipeline

# test BSCAMPP
def test_bscampp_pipeline():
    res = bscampp_pipeline(dry_run=True)
    assert res == True

    # remove bscampp_output that's created
    if os.path.isdir('bscampp_output'):
        os.rmdir('bscampp_output')

# test SCAMPP (almost the same as BSCAMPP)
def test_scampp_pipeline():
    res = scampp_pipeline(dry_run=True)
    assert res == True

    # remove scampp_output that's created
    if os.path.isdir('scampp_output'):
        os.rmdir('scampp_output')
